// psfex.cpp: provides access to PSFEx models
// originally written by Daniel Gruen, 2010-2013

/*
    This file is part of the PSFEx library.

    The PSFEx library is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSFEx library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the PSFEx library.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "psfex.h"

using namespace std;

psfex::psfex(char *filename)
	// constructor, reads PSFEx PSF model from FITS format as generated by PSFEx or ASCII file as generated by psfextract.py
	{
		  // initialize
		 
		  loaded=accepted=0; chi2=1.e30;
		  polnaxis=0;
		  psf_samp=0;
		  
		  n12=n123=0;
	  
		  // do a basic file check
		  ifstream in(filename);
		  if(!in.good()) {
		    std::cerr << "Could not open PSFEx model file " << filename;
		    throw PSFEX_CANNOT_OPEN;
		  }
		  
		  // check for the extensions and
		  // initialize either from ASCII or FITS
		  string tmp(filename);
		  if (tmp.rfind(".psf") != -1 || tmp.rfind(".fits") != -1 || tmp.rfind(".fit") != -1 ) {
		    open_from_fits(filename);
		  } else {
		    open_from_ascii(filename);
		  }
		  
		  // some useful numbers
		  n12 = psfaxis[0]*psfaxis[1];
		  n123 = n12*psfaxis[2];
		  
		  // prepare cache
		  cache.resize(n12);
		  xcached = ycached = 1.e30;
		  
	}

void psfex::open_from_ascii(char *filename)
	{
		std::ifstream in(filename);
		string buf;
		bool headerread=false; // flag set true once header has been read completely
		bool polnaxisread=false;
		bool polngrpread=false;
		bool psfnaxisread=false;
		do
		{
			in >> buf;
			if(!buf.compare("LOADED"))
			{
				in >> loaded;
			}
			else if(!buf.compare("ACCEPTED"))
			{
				in >> accepted;
			}
			else if(!buf.compare("CHI2"))
			{
				in >> chi2;
			}
			else if(!buf.compare("POLNAXIS"))
			{
				in >> polnaxis;
				if(polnaxis && polnaxis!=2)
				{
					std::cerr << "This psf file uses not exactly two parameters for polynomial interpolation. Interpolation will not work!" << std::endl;
					throw PSFEX_UNKNOWN_FORMAT;
				}
				polgrp.resize(polnaxis, 0);
				polname.resize(polnaxis, "UNKNOWN");
				polzero.resize(polnaxis, 0.);
				polscal.resize(polnaxis, 1.);
				polnaxisread=true;
			}
			else if(!buf.compare(0,6,"POLGRP"))
			{
				int i = atoi(buf.substr(6).c_str())-1;
				if(polnaxisread && i<polnaxis)
				{
					in >> polgrp[i];
				}
				else
				{
					std::cerr << "syntax error: reading axis parameter POLGRP without proper polnaxis" << std::endl;
					throw PSFEX_UNKNOWN_FORMAT;
				}
			}
			else if(!buf.compare(0,7,"POLNAME"))
			{
				int i = atoi(buf.substr(7).c_str())-1;
				if(polnaxisread && i<polnaxis)
				{
					in >> polname[i];
				}
				else
				{
					std::cerr << "syntax error: reading axis parameter POLNAME without proper polnaxis" << std::endl;
					throw PSFEX_UNKNOWN_FORMAT;
				}
			}
			else if(!buf.compare(0,7,"POLZERO"))
			{
				int i = atoi(buf.substr(7).c_str())-1;
				if(polnaxisread && i<polnaxis)
				{
					in >> polzero[i];
				}
				else
				{
					std::cerr << "syntax error: reading axis parameter POLZERO without proper polnaxis" << std::endl;
					throw PSFEX_UNKNOWN_FORMAT;
				}
			}
			else if(!buf.compare(0,7,"POLSCAL"))
			{
				int i = atoi(buf.substr(7).c_str())-1;
				if(polnaxisread && i<polnaxis)
				{
					in >> polscal[i];
				}
				else
				{
					std::cerr << "syntax error: reading axis parameter POLSCAL without proper polnaxis" << std::endl;
					throw PSFEX_UNKNOWN_FORMAT;
				}
			}

			else if(!buf.compare("POLNGRP"))
			{
				in >> polngrp;
				if(polngrp>1)
				{
					std::cerr << "this psf file uses more than one polynomial group. Interpolation will not work!" << std::endl;
					throw PSFEX_UNKNOWN_FORMAT;
				}
				poldeg.resize((polngrp>0)?polngrp:1, 0); // make poldeg[0]=0 even if polngrp is 0
				polngrpread = true;

			}
			else if(!buf.compare(0,6,"POLDEG"))
			{
				int i = atoi(buf.substr(6).c_str())-1;
				if(polngrpread && i<polngrp)
				{
					in >> poldeg[i];
				}
				else
				{
					std::cerr << "syntax error: reading group parameter POLDEG without proper polngrp" << std::endl;
					throw PSFEX_UNKNOWN_FORMAT;
				}
			}
			else if(!buf.compare("PSF_FWHM"))
			{
				in >> psf_fwhm;
			}
			else if(!buf.compare("PSF_SAMP"))
			{
				in >> psf_samp;
			}
			else if(!buf.compare("CHI2"))
			{
				in >> chi2;
			}
			else if(!buf.compare("PSFNAXIS"))
			{
				in >> psfnaxis;
				if(psfnaxis!=3)
				{
					 std::cerr << "syntax error: PSFNAXIS must be 3" << std::endl;
					 throw PSFEX_UNKNOWN_FORMAT;
				}
				psfaxis.resize(psfnaxis, 0);
				psfnaxisread=true;
			}
			else if(!buf.compare(0,7,"PSFAXIS"))
			{
				int i = atoi(buf.substr(7).c_str())-1;
				if(psfnaxisread && i<psfnaxis)
				{
					in >> psfaxis[i];
				}
				else
				{
					std::cerr << "syntax error: reading psf axis parameter PSFAXIS without proper psfnaxis" << std::endl;
					throw PSFEX_UNKNOWN_FORMAT;
				}
			}
			else if(!buf.compare("DATA"))
			{
				headerread=true;
			}
			else
			{
				std::cerr << "ignoring unknown header word " << buf << std::endl;
			}
		} while(!headerread && !in.eof());

		if(!headerread || !psfnaxisread || !polngrpread || !polnaxisread || psfnaxis!=3 || psfaxis[0]<=0 || psfaxis[1]<=0 || psfaxis[2]<=0)
		{
			std::cerr << "syntax error in header" << std::endl;
			throw PSFEX_UNKNOWN_FORMAT;
		}

		data.resize(1);
		data[0].resize(psfaxis[0]*psfaxis[1]*psfaxis[2],0);

		int ri=0; // running index
		for(int k=0; k<psfaxis[2]; k++)
		{
			double sum = 0.0;
			for(int j=0; j<psfaxis[1]; j++)
			{
				for(int i=0; i<psfaxis[0]; i++)
				{
					in >> data[0][ri];
					ri++;
					// TODO: check file is not finished already
				}
			}
		}
	}
	
void psfex::open_from_fits(char *filename)
	{
		auto_ptr<CCfits::FITS> pInfile(0);

		try {
			// open the fits table and go to the right extension
			pInfile.reset(new CCfits::FITS(filename,CCfits::Read, "PSF_DATA"));
		}
		catch (CCfits::FITS::CantOpen &fitsError) {
			std::cerr << "Error opening the file with message: " << fitsError.message();
			throw PSFEX_FITS_ERROR;
		}
		catch (CCfits::FITS::NoSuchHDU &fitsError) {
			std::cerr << "Error going to the HDU: " << fitsError.message();
		}
		catch (CCfits::FitsException &fitsError) {
			std::cerr << "Error in FITS: " << fitsError.message();
		}

		// go to the right extension
		CCfits::ExtHDU& table = pInfile->extension("PSF_DATA");

		// read the keyword "POLNAXIS"
		// dimension some arrays
		
		readTableLKeyword <int> (table, string("POLNAXIS"), polnaxis);
		if(polnaxis && polnaxis!=2) {
			std::cerr << "This psf file uses not exactly two parameters for polynomial interpolation. Interpolation will not work!" << std::endl;
			throw PSFEX_UNKNOWN_FORMAT;
		}
		polgrp.resize(polnaxis, 0);
		polname.resize(polnaxis, "UNKNOWN");
		polzero.resize(polnaxis, 0.);
		polscal.resize(polnaxis, 1.);

		// read the indexed keywords "POLGRP", "POLNAME", "POLZERO" and "POLSCAL" to an array
		for (int index=0; index < polnaxis; index++){
			// convert the index to a keyword appendix
			string strIndex = static_cast<ostringstream*>( &(ostringstream() << index+1) )->str();

			// read the indexed keyword "POLGRP"
			readTableLKeyword <int> (table, string("POLGRP")+strIndex, polgrp[index]);

			// read the indexed keyword "POLNAME"
			readTableLKeyword <string> (table, string("POLNAME")+strIndex, polname[index]);

			// read the indexed keyword "POLZERO"
			readTableLKeyword <double> (table, string("POLZERO")+strIndex, polzero[index]);

			// read the indexed keyword "POLSCAL"
			readTableLKeyword <double> (table, string("POLSCAL")+strIndex, polscal[index]);
		}

		// read the keyword "POLNGRP"
		// dimension an array
		readTableLKeyword <int> (table, string("POLNGRP"), polngrp);
		
		if (polngrp>1) 
		{
			std::cerr << "this psf file uses more than one polynomial group. Interpolation will not work!" << std::endl;
			throw PSFEX_UNKNOWN_FORMAT;
		}

		poldeg.resize((polngrp>0)?polngrp:1, 0); // make poldeg[0]=0 even if polngrp is 0

		// read the indexed keyword "POLDEG" to an array
		for (int index=0; index < polngrp; index++){
			// convert the index to a keyword appendix
			string strIndex = static_cast<ostringstream*>( &(ostringstream() << index+1) )->str();

			// read one indexed keyword
			readTableLKeyword <int> (table, string("POLDEG")+strIndex, poldeg[index]);
		}

		// read the keyword "PSF_FWHM"
		readTableLKeyword <double> (table, string("PSF_FWHM"), psf_fwhm);

		// read the keyword "PSF_SAMP"
		readTableLKeyword <double> (table, string("PSF_SAMP"), psf_samp);

		// read the keyword "CHI2"
		readTableLKeyword <double> (table, string("CHI2"), chi2);

		// read the keyword "PSFNAXIS"
		// dimension an array
		
		readTableLKeyword <int> (table, string("PSFNAXIS"), psfnaxis);

		if(psfnaxis!=3)
		{
			 std::cerr << "syntax error: PSFNAXIS must be 3" << std::endl;
			 throw PSFEX_UNKNOWN_FORMAT;
		}
		
		psfaxis.resize(psfnaxis, 0);

		// read the indexed keyword "PSFAXIS" to an array
		for (int index=0; index < psfnaxis; index++){
			// convert the index to a keyword appendix
			string strIndex = static_cast<ostringstream*>( &(ostringstream() << index+1) )->str();

			// read one indexed keyword
			readTableLKeyword <int> (table, string("PSFAXIS")+strIndex, psfaxis[index]);
		}

		// read the data 
		// check for some typical errors
		try {
			table.column("PSF_MASK").readArrays( data, 0, 0);
		}
		catch (CCfits::Table::NoSuchColumn &fitsError) {
			std::cerr << "Error reading column PSF_MASK: " << fitsError.message() << std::endl;
			throw PSFEX_DATA_ERROR;
		}
		catch (CCfits::Column::InvalidRowNumber &fitsError)
		{
			std::cerr << "Error reading a table row: " << fitsError.message() << std::endl;
			throw PSFEX_DATA_ERROR;
		}
		catch (CCfits::FitsException &fitsError)
		{
			std::cerr << "Error in FITS: " << fitsError.message() << std::endl;
			throw PSFEX_DATA_ERROR;
		}
		
		// data[0] is a valarray with all pixels, small steps are x, medium steps are y, large steps are components
		// use data[0][index(k, x, y)] to access this
		
	}

psfex::~psfex()
	{
	}
	
int psfex::index(const int k, const int x, const int y) const
	// return index of pixel in data array
	// for component k at position x,y (all running from 0 to their maximum)
	{
	  return x+psfaxis[0]*y+psfaxis[0]*psfaxis[1]*k; 
	}
	
void psfex::cache_position(double x, double y) // cache postage stamp image at position
	{
		if(x==xcached && y==ycached) return; // got this position cached already
		
		double dx = 0; 
		double dy = 0; 
		
		if(poldeg[0]>0) {
		  dx=(x-polzero[0])/polscal[0];
		  dy=(y-polzero[1])/polscal[1];
		}
		
		for(int i=0; i<n12; i++)
		  {
		      cache[i] = data[0][i]; // constant component
		  }
		
	  	for(int n=1; n<=poldeg[0]; n++) // iterate through all orders from 0 to n
		{
		  for(int ny=0; ny<=n; ny++) // orders in y
		  {
		    int nx = n-ny; // order in x so x^(nx)*y^(ny) is of order n=nx+ny
		    int n3 = nx+ny*(poldeg[0]+1)-(ny*(ny-1))/2;
		    double nf = pow(dx,nx)*pow(dy,ny);
		    
		    for(int i=0; i<n12; i++)
		    {
			cache[i] += nf*data[0][i+n12*n];
		    }
		  }
		}
		
		xcached=x;
		ycached=y;
	}
	
vector<double> psfex::centroid(double x, double y)
	{
	    cache_position(x,y);
	    double f=0.;
	    double fx=0.;
	    double fy=0.;
	    
	    int ri=0;
	    for(int i=0; i<psfaxis[0]; i++)
	    {
	      for(int j=0; j<psfaxis[1]; j++)
	      {
		f+=cache[ri];
		fx+=i*cache[ri];
		fy+=j*cache[ri];
		ri++;
	      }
	    }
	    
	    vector<double> position;
	    //std::cerr << "x centroid in psf_samp size pixels is: " << fx/f << std::endl;
	    position.push_back(fx/f*psf_samp);
	    position.push_back(fy/f*psf_samp);
	    return position;
	}
	
double psfex::pixel_raw(int i, int j, double x, double y, bool cached)
	// i-th psf (!! sampled in psf_samp size pixels, so directly from the PSFEx model without interpolation !!) pixel in x direction,
	// j-th psf pixel in y direction, model evaluated at (x,y) central coordinate of psf
	// caveat: only works when there's just one polngrp
	{ 
		if(i>=0 && j>=0 && i<psfaxis[0] && j<psfaxis[1])
		{
			int i12=i+j*psfaxis[1];
		  	if(cached) {
			  cache_position(x,y);
			  return cache[i12];
			}

			double dx = 0; 
			double dy = 0; 
			
			if(poldeg[0]>0) {
			  dx=(x-polzero[0])/polscal[0];
			  dy=(y-polzero[1])/polscal[1];
			}
			
			double result = data[0][i12];  // 0th order

			for(int n=1; n<=poldeg[0]; n++) // iterate through all orders from 1 to n (zero done above already)
			{
			  for(int ny=0; ny<=n; ny++) // orders in y
			  {
				int nx = n-ny; // order in x so x^(nx)*y^(ny) is of order n=nx+ny
				result += pow(dx,nx)*pow(dy,ny)*data[0][i12+n12*(nx+ny*(poldeg[0]+1)-(ny*(ny-1))/2)];
		          }
			}

			return result;
		}

		std::cerr << "requesting pixel outside valid range " << i << " " << j << std::endl;
		throw PSFEX_OUTSIDE;

		return 42; // never happens
	}

double psfex::pixel_lanczos(double relx, double rely, double centerx, double centery, int lanczos_order)
	// image sampled psf pixel at position relx,rely relative to the psf center at centerx,centery (all in image coordinates)
	// using lanczos interpolation of order lanczos_order (best use 3 or 4)
	// center of PSF is at relx=rely=0, but there can be offsets intrinsic to the model (see psf::centroid(), which should be used to correct that)
	{
		double relxp = relx/psf_samp;
		double relyp = rely/psf_samp;

		if(fabs(relxp)>(psfaxis[0]-1)/2.-lanczos_order || fabs(relyp)>(psfaxis[1]-1)/2.-lanczos_order )
		{
			// out of bounds, assume pixel is 0
			return 0;
		}

		double sum = 0.;
		for(int i=0; i<psfaxis[0]; i++) // x coordinate on psf_sample scale
		for(int j=0; j<psfaxis[1]; j++) // y coordinate on psf_sample scale
		{
			double dx = fabs(( double(i)-double(psfaxis[0]-1)/2. ) - relxp);
			double dy = fabs(( double(j)-double(psfaxis[1]-1)/2. ) - relyp);

			if(dx<lanczos_order && dy<lanczos_order)
			{
				double interpolant = sinc(dx)*sinc(dx/lanczos_order)*sinc(dy)*sinc(dy/lanczos_order);
				double value = pixel_raw(i,j,centerx,centery);
				sum += value*interpolant;
			}
		}
		return sum/(pow(psf_samp,2.));
	}
	
double psfex::sampling() const {
  return psf_samp;
}

template <class TT>
void psfex::readTableLKeyword(CCfits::ExtHDU& table, string kWord, TT& parT)
	{
		// read the keyword
		try {
			table.readKey(string(kWord), parT);
		}
		catch (CCfits::HDU::NoSuchKeyword &fitsError) {
			std::cerr << "Error reading keyword " <<  kWord  << ": " << fitsError.message() << std::endl;
			throw PSFEX_UNKNOWN_FORMAT;
		}
	}
	
double psfex::maxsize(double lanczos_order) const
	// distance from center out to which model is available (image-size pix)
	{
		return ( (psfaxis[0]-1)/2.-lanczos_order ) * psf_samp;
	}
