// psfex.h: provides access to PSFEx models
// originally written by Daniel Gruen, 2010-2013

/*
    This file is part of the PSFEx library.

    The PSFEx library is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The PSFEx library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the PSFEx library.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include <iostream>
#include <cstdlib>
#include <vector>
#include <valarray>
#include <fstream>
#include <cstring>
#include <sstream>
#include <CCfits/CCfits>

// error codes
#define PSFEX_CANNOT_OPEN    1  // file could not be opened
#define PSFEX_UNKNOWN_FORMAT 2  // class in unable to correctly read and/or interpret PSFEx model; either the file is wrong or is uses a feature not yet implemented here
#define PSFEX_FITS_ERROR     3  // error reading FITS file
#define PSFEX_DATA_ERROR     4  // data is incomplete or malformatted
#define PSFEX_OUTSIDE        5  // requesting pixel outside valid range

class psfex
{
   public:

	psfex(char *filename);
	// constructor, reads PSFEx PSF model from FITS format as generated by PSFEx or ASCII file
	
	~psfex();
	
	std::vector<double> centroid(double x, double y);
	// returns centroid (in sampling() size pixels) of PSFEx model at position x,y in the FP; also caches postage stamp at that position

	// access these variables if interested in how well the model was fit
	int loaded, accepted;
	double chi2;
	
	//////// image access routines

	double sampling() const; // return sampling scale of PSFEx model; e.g. 0.5 means that model pixels are smaller than image pixels

	// raw
	double pixel_raw(int i, int j, double x, double y, bool cached=true);
	// i-th psf (!! sampled in sampling() size pixels, so directly from the PSFEx model without interpolation !!) pixel in x direction,
	// j-th psf pixel in y direction, 
	// model evaluated at (x,y) central FP coordinate of psf
	// caveat: no normalization applied!
	
	// interpolated to pixel scale
	double pixel_lanczos(double relx, double rely, double centerx, double centery, int lanczos_order=4);
	// image sampled psf pixel at position relx,rely relative to the psf center at centerx,centery (image coordinates)
	// using lanczos interpolation of order lanczos_order (use 3 or 4)
	// center of PSF is at relx=rely=0, but there can be offsets intrinsic to the model (see psfex::centroid())
	// caveat: no normalization applied

	double maxsize(double lanczos_order=4) const;
	// distance from center out to which model is available (image-size pix)

  private:
	void open_from_ascii(char *filename);
	// reads ASCII format, called by constructor

	void open_from_fits(char *filename);
	// reads FITS format PSFEx files, called by constructor
	
	int index(const int k, const int x, const int y) const;
	// return index of pixel in data array
	// for component k at position x,y (all running from 0 to their maximum)
	// internal use only

	void cache_position(double x, double y); 
	// cache postage stamp image at position 
	// internal use only

	template <class TT>
	void readTableLKeyword(CCfits::ExtHDU& table, string kWord, TT& parT);

	static double sinc(double x) // normalized sinc function, see PSFEx manual
	{
		if(x<1e-5 && x>-1e-5)
			return 1.;
		return sin(x*M_PI)/(x*M_PI);
	}

	
	//////////// internal variables

	int psfnaxis;		     // dimensionality of PSF datacube (should be 3 always)
	std::vector<int> psfaxis;    // psfnaxis-dimensional
	int n12, n123;
	
	int polnaxis;		// number of variables in the polynomial; usually 2 (X and Y)
	std::vector<int> polgrp; 
	std::vector<double> polzero;
	std::vector<double> polscal;
	std::vector<string> polname; // all polnaxis-dimensional
		
	int polngrp;		// number of 'polynomial groups'; only 1 is implemented
	std::vector<int> poldeg;     // polngrp-dimensional
	
	double psf_samp; double psf_fwhm;
	
	std::vector <valarray <double> > data; // a vector just so CCfits likes it, the 0th element holds the PSF data
	std::valarray<double> cache; 
	double xcached, ycached;
	
};